// Creative Commons Attribution-Share Alike 3.0 Unported Licence
// http://creativecommons.org/licenses/by-sa/3.0/

var CGD = window.CGD || {};

CGD.TIME = CGD.TIME || {};
// Management routines (start, stop, etc.) around the setTimeout system
CGD.TIME.ticker = function (call, timeout) {
  // The local 'my' is used in preference to 'this' because the 
  //   routines can be detached from their object - in particular the timer callback.
  var my = {
    handle: null, // setTimeout return value
    // set the period between firings; resets the timer if it's running,
    //   but leaves it alone if it's not.
    configure: function(period) {
      if (isNaN(period) || period < 10) {
        throw new RangeError('Timer out of range - ' + period);
      }
      timeout = period;
      if (my.running()) {
        my.stop();
        my.start();
      }
    },
    // sets the time, and will start the timer if it's not running.
    restart: function(period) {
      timeout = period;
      my.stop();
      my.start();
    },
    start: function() {
      if (my.handle == null) {
        my.handle = setTimeout(my.ontimeout,
          // implements synchronized events; setInterval drifts
          Math.ceil(timeout - (new Date().getTime() % timeout)));
      }
    },
    // set a one-time delay.
    forestall: function(delay) {
      my.stop();
      my.handle = setTimeout(my.ontimeout, delay);
    },
    stop: function () {
      if (my.handle != null) {
        clearTimeout(my.handle);
        my.handle = null;
      }
    },
    running: function () {
      return (my.handle != null);
    },
    toString: function () {
      return '[Ticker ' + call + ' /' + timeout + ']';
    },
    // the timer event.
    ontimeout: function() {
      my.stop();
      //CGD.DEBUG.p(new Date().getTime() % timeout);
      call();
      // Note that start won't be called if there is an exception in the
      //  target routine - This very much intentional, since browsers don't seem
      //  to like a flood of exceptions much ;^)
      my.start();
    }
  };
  return my;
};

// Convenience version of CGD.JS.flag for time based events.
CGD.TIME.flag = function(timeout) {
  var my = CGD.JS.flag();
  setTimeout(my, timeout);
  return my;
};

// Control the call rate of a function.
//  Note that this version makes no attempt to handle parameters
//  It always induces a delay, which makes the state tracking much easier.
//    This makes it only useful for small time periods, however.
//    Just clearing a flag after the timeout won't work, because we want
//    to do the the routine at least once if was called during the wait period.
CGD.TIME.throttle = function(call, timeout) {
  var my = function() {
    if (!my.handle) {
      my.handle = setTimeout(my.rightNow, timeout);
    }
  };
  my.rightNow = function() {
    if (my.handle) {
      clearTimeout(my.handle);
      my.handle = null;
    }
    call();
  };
  return my;
};

CGD.TIME.UNIT = {
  // conveience units
  MS_S: 1000,
  S_M: 60,
  M_H: 60,
  H_D: 24,
  MS_D: 1000 *  60 * 60 * 24,
  D_W: 7,
  D_LM: 29.530589,
  D_SY: 365.24219,
  Y_G: 20,
  G_L: 4,
  Y_C: 100,
  D_TM: 20,
  D_TR: 13,
  // kin - day
  D_U: 20,
  U_T: 18,
  T_KAT: 20,
  KAT_BAKT: 20,
  BAKT_PICT: 20,
  MAYAN_EPOCH: 2440587 - 584282,
  // Mayan offset and related calculations greatly assisted by reference
  //   to Fourimlab's calendar converter.
  // http://www.fourmilab.ch/documents/calendar/
  nameFor: function (ms) {
    try {
      return CGD.OBJECT.keyWhich(CGD.TIME.UNIT.MS.per, CGD.OBJECT.equals(ms));
    } catch (e) {
      CGD.DEBUG.p(e);
      return '??';
    }
    return 'huh?';
  },
  forName: function (name) {
    return CGD.TIME.UNIT.MS.per[name] || 0;
  }
};

with (CGD.TIME.UNIT) {
  // MS and DAYS are lookup tables for turning units back into names
  // the separation of DAYS is basically to avoid writing MS_D a lot.
  CGD.TIME.UNIT.MS = {
    per: {
      ms: 1,
      tenth: 100,
      second: MS_S,
      minute: MS_S * S_M,
      hour: MS_S * S_M * M_H,
      day: MS_D
    }
  };
  
  CGD.TIME.UNIT.DAYS = {
    per: {
      day: 1,
      week: D_W,
      moon: D_LM,
      '~month': D_SY/12,
      year: D_SY,
      generation: D_SY * Y_G,
      lifetime: D_SY * Y_G * G_L,
      century: D_SY * Y_C,
      kin: 1,
      uinal: D_U,
      tun: D_U * U_T,
      katun: D_U * U_T * T_KAT,
      baktun: D_U * U_T * T_KAT * KAT_BAKT,
      pictun: D_U * U_T * T_KAT * KAT_BAKT * BAKT_PICT,
      'tzolk\'in month': D_TM,
      'tzolk\'in trecena': D_TR,
      'tzolk\'in year': D_TM * D_TR
    }
  };

  CGD.OBJECT.forEach(DAYS.per, function(value, key) {
    MS.per[key] = value * MS_D;
  });
}